Require Import Vbase Relations ClassicalDescription ClassicalChoice.
Require Import List Permutation Sorting extralib. 
Require Import c11 exps.
Require Import GpsSepAlg GpsModel GpsModelLemmas GpsLogic.

Local Open Scope string_scope.

Set Implicit Arguments.

(** Separation algebra for tokens *)

Program Definition sa_tok :=
  {| salg := bool
   ; salg_emp := false
   ; salg_plus x y := if x && y then None else Some (x || y)
   ; salg_strip x := false
  |}.
Solve Obligations with
  (ins; repeat match goal with b : bool |- _ => destruct b; ins end; vauto).

(** Product construction for separation algebras *)

Program Definition sa_pair (sa1 sa2 : salg_mod) :=
  {| salg := salg sa1 * salg sa2 
   ; salg_emp := (salg_emp sa1, salg_emp sa2)
   ; salg_plus x y := 
       match salg_plus sa1 (fst x) (fst y) return _ with None => None | Some a =>
       match salg_plus sa2 (snd x) (snd y) return _ with None => None | Some b =>
       Some (a, b) end end
   ; salg_strip x := (salg_strip sa1 (fst x), salg_strip sa2 (snd x))
  |}.
Next Obligation. by ins; rewrite (salg_plusC sa1), (salg_plusC sa2). Qed.
Next Obligation. by ins; rewrite (salg_plusI sa1), (salg_plusI sa2); destruct x. Qed.
Next Obligation. 
ins; desf_asm; ins;
match goal with H1 : _ , H2 : _ |- _ => destruct (salg_plusA sa1 _ _ H1 _ H2) as (m & A & ?) end;
match goal with H1 : _ , H2 : _ |- _ => destruct (salg_plusA sa2 _ _ H1 _ H2) as (n & B & ?) end;
eexists (m,n); ins; desf. 
Qed.
Next Obligation. 
ins; desf; ins; vauto; right;
try match goal with H1 : _ , H2 : _ |- _ => destruct (salg_plusA sa1 _ _ H1 _ H2) as (m & A & ?) end;
try match goal with H1 : _ , H2 : _ |- _ => destruct (salg_plusA sa2 _ _ H1 _ H2) as (n & B & ?) end;
try match goal with H1 : _ , H2 : _ |- _ => destruct (salg_plusA2 sa1 _ _ H1 _ H2) as [|(m & A & ?)] end;
try match goal with H1 : _ , H2 : _ |- _ => destruct (salg_plusA2 sa2 _ _ H1 _ H2) as [|(n & B & ?)] end;
try eexists (m,n); ins; desf. 
eexists (s,s0); split; ins; desf.
Qed.
Next Obligation. ins; desf_asm; destruct g; f_equal; eauto using salg_eq_empD. Qed.
Next Obligation. by ins; rewrite !salg_strip_emp. Qed.
Next Obligation. 
by ins; desf_asm; ins; repeat match goal with H : _ |- _ => rewrite (salg_strip_plus _ _ _ H) in * end.
Qed.
Next Obligation. by ins; rewrite !salg_plus_strip_idemp; destruct x. Qed.
Next Obligation. by destruct y; ins; desf; f_equal; eauto using salg_strip_max. Qed.
Next Obligation.
by ins; desf_asm;
   match goal with H1 : _ , H2 : _ |- _ => rewrite (salg_pcancel sa1 _ _ _ H1 H2) end;
   match goal with H1 : _ , H2 : _ |- _ => rewrite (salg_pcancel sa2 _ _ _ H1 H2) end. 
Qed.

Definition tok := Some (existT [eta salg] sa_tok true).

Definition Tok a := Rgres (gres_one a tok). 

Lemma star_tok_tok a :
  riff (Rstar (Tok a) (Tok a)) Rfalse.
Proof.
  intro; unfold Tok; rewrite Rstar_gres1; split; ins; desf; clear - H.
  unfold gres_plus, gres_one in *; des_eqrefl; ins; desf.
  inv_lplus a; desf; desf; ins; desf; desf; ins.
  rewrite <- eq_rect_eq in *; ins.
Qed.

Lemma new_tok IE P e QQ :
  (forall g, triple IE (Rstar P (Tok g)) e QQ) ->
  triple IE P e QQ.
Proof.
  ins; eapply rule_conseq;
    try eapply rule_exists with (PP := fun g => Rstar P (Tok g)); ins; 
    eauto using gimp_refl; try rewrite <- app_nil_end; ins.
  rewrite <- star_exists, <- (Rstar_true P) at 1.
  eapply gimp_frame_l, (gimp_gintro _ tok).
Qed.


